''***************************************
''*  VGA High-Res Text Driver v1.0      *
''*  Author: Chip Gracey                *
''*  Copyright (c) 2006 Parallax, Inc.  *
''*  See end of file for terms of use.  *
''***************************************
''
'' This object generates a 1024x768 VGA signal which contains 128 columns x 64
'' rows of 8x12 characters. Each row can have a unique forground/background
'' color combination and each character can be inversed. There are also two
'' cursors which can be independently controlled (ie. mouse and keyboard). A
'' sync indicator signals each time the screen is refreshed (you may ignore).
''
'' You must provide buffers for the screen, colors, cursors, and sync. Once
'' started, all interfacing is done via memory. To this object, all buffers are
'' read-only, with the exception of the sync indicator which gets written with
'' -1. You may freely write all buffers to affect screen appearance. Have fun!
''

CON

{
' 1024 x 768 @ 57Hz settings: 128 x 48 characters

  hp = 1024     'horizontal pixels
  vp = 768      'vertical pixels
  hf = 16       'horizontal front porch pixels
  hs = 96       'horizontal sync pixels
  hb = 176      'horizontal back porch pixels
  vf = 1        'vertical front porch lines
  vs = 3        'vertical sync lines
  vb = 28       'vertical back porch lines
  hn = 1        'horizontal normal sync state (0|1)
  vn = 1        'vertical normal sync state (0|1)
  pr = 60       'pixel rate in MHz at 80MHz system clock (5MHz granularity)
}
'{
' 800 x 600 @ 75Hz settings: 100 x 37 characters

  hp = 800      'horizontal pixels
  vp = 600      'vertical pixels
  hf = 40       'horizontal front porch pixels
  hs = 128      'horizontal sync pixels
  hb = 88       'horizontal back porch pixels
  vf = 5        'vertical front porch lines
  vs = 4        'vertical sync lines
  vb = 27       'vertical back porch lines
  hn = 0        'horizontal normal sync state (0|1)
  vn = 0        'vertical normal sync state (0|1)
  pr = 50       'pixel rate in MHz at 80MHz system clock (5MHz granularity)
'}
{
' 640 x 480 @ 69Hz settings: 80 x 30 characters

  hp = 640      'horizontal pixels
  vp = 480      'vertical pixels
  hf = 24       'horizontal front porch pixels
  hs = 40       'horizontal sync pixels
  hb = 128      'horizontal back porch pixels
  vf = 9        'vertical front porch lines
  vs = 3        'vertical sync lines
  vb = 28       'vertical back porch lines
  hn = 1        'horizontal normal sync state (0|1)
  vn = 1        'vertical normal sync state (0|1)
  pr = 30       'pixel rate in MHz at 80MHz system clock (5MHz granularity)
}

' columns and rows

  cols = hp / 16
  rows = vp / 16


VAR long cog[2]

PUB start(BasePin, ScreenPtr, ColorPtr, CursorPtr, SyncPtr) : okay | i, j

'' Start VGA driver - starts two COGs
'' returns false if two COGs not available
''
''     BasePin = VGA starting pin (0, 8, 16, 24, etc.)
''
''   ScreenPtr = Pointer to 8,192 bytes containing ASCII codes for each of the
''               128x64 screen characters. Each byte's top bit controls color
''               inversion while the lower seven bits provide the ASCII code.
''               Screen memory is arranged left-to-right, top-to-bottom.
''
''               screen byte example: %1_1000001 = inverse "A"
''
''    ColorPtr = Pointer to 64 words which define the foreground and background
''               colors for each row. The lower byte of each word contains the
''               foreground RGB data for that row, while the upper byte
''               contains the background RGB data. The RGB data in each byte is
''               arranged as %RRGGBB00 (4 levels each).
''
''               color word example: %%0020_3300 = gold on blue
''
''   CursorPtr = Pointer to 6 bytes which control the cursors:
''
''               bytes 0,1,2: X, Y, and MODE of cursor 0
''               bytes 3,4,5: X, Y, and MODE of cursor 1
''
''               X and Y are in terms of screen characters
''               (left-to-right, top-to-bottom)
''
''               MODE uses three bottom bits:
''
''                      %x00 = cursor off
''                      %x01 = cursor on
''                      %x10 = cursor on, blink slow
''                      %x11 = cursor on, blink fast
''                      %0xx = cursor is solid block
''                      %1xx = cursor is underscore
''
''               cursor example: 127, 63, %010 = blinking block in lower-right
''
''     SyncPtr = Pointer to long which gets written with -1 upon each screen
''               refresh. May be used to time writes/scrolls, so that chopiness
''               can be avoided. You must clear it each time if you want to see
''               it re-trigger.

  'if driver is already running, stop it
  stop

  'implant pin settings
  reg_vcfg := $200000FF + (BasePin & %111000) << 6
  i := $FF << (BasePin & %011000)
  j := BasePin & %100000 == 0
  reg_dira := i & j
  reg_dirb := i & !j

  'implant CNT value to sync COGs to
  sync_cnt := cnt + $10000

  'implant pointers
  longmove(@screen_base, @ScreenPtr, 3)
  font_base := @font

  'implant unique settings and launch first COG
  vf_lines.byte := vf
  vb_lines.byte := vb
  font_quarter := 1
  cog[1] := cognew(@d0, SyncPtr) + 1

  'allow time for first COG to launch
  waitcnt($2000 + cnt)

  'differentiate settings and launch second COG
  vf_lines.byte := vf+4
  vb_lines.byte := vb-4
  font_quarter := 0
  cog[0] := cognew(@d0, SyncPtr) + 1

  'if both COGs launched, return true
  if cog[0] and cog[1]
    return true

  'else, stop any launched COG and return false
  else
    stop


PUB stop | i

'' Stop VGA driver - frees two COGs

  repeat i from 0 to 1
    if cog[i]
      cogstop(cog[i]~ - 1)


CON

  #1, scanbuff[128], scancode[128*2-1+3], maincode      'enumerate COG RAM usage

  main_size = $1F0 - maincode                           'size of main program

  hv_inactive = (hn << 1 + vn) * $0101                  'H,V inactive states


DAT

'*****************************************************
'* Assembly language VGA high-resolution text driver *
'*****************************************************

' This program runs concurrently in two different COGs.
'
' Each COG's program has different values implanted for front-porch lines and
' back-porch lines which surround the vertical sync pulse lines. This allows
' timed interleaving of their active display signals during the visible portion
' of the field scan. Also, they are differentiated so that one COG displays
' even four-line groups while the other COG displays odd four-line groups.
'
' These COGs are launched in the PUB 'start' and are programmed to synchronize
' their PLL-driven video circuits so that they can alternately prepare sets of
' four scan lines and then display them. The COG-to-COG switchover is seemless
' due to two things: exact synchronization of the two video circuits and the
' fact that all COGs' driven output states get OR'd together, allowing one COG
' to output lows during its preparatory state while the other COG effectively
' drives the pins to create the visible and sync portions of its scan lines.
' During non-visible scan lines, both COGs output together in unison.
'
' COG RAM usage:  $000      = d0 - used to inc destination fields for indirection
'                 $001-$080 = scanbuff - longs which hold 4 scan lines
'                 $081-$182 = scancode - stacked WAITVID/SHR for fast display
'                 $183-$1EF = maincode - main program loop which drives display

                        org    0                         'set origin to $000 for start of program

d0                      long    1 << 9                  'd0 always resides here at $000, executes as NOP


' Initialization code and data - after execution, space gets reused as scanbuff

                        'Move main program into maincode area

:move                   mov     $1EF,main_begin+main_size-1
                        sub     :move,d0s0              '(do reverse move to avoid overwrite)
                        djnz    main_ctr,#:move

                        'Build scanbuff display routine into scancode

:waitvid                mov     scancode+0,i0           'org     scancode
:shr                    mov     scancode+1,i1           'waitvid color,scanbuff+0
                        add     :waitvid,d1             'shr     scanbuff+0,#8
                        add     :shr,d1                 'waitvid color,scanbuff+1
                        add     i0,#1                   'shr     scanbuff+1,#8
                        add     i1,d0                   '...
                        djnz    scan_ctr,#:waitvid      'waitvid color,scanbuff+cols-1

                        mov     scancode+cols*2-1,i2    'mov     vscl,#hf
                        mov     scancode+cols*2+0,i3    'waitvid hvsync,#0
                        mov     scancode+cols*2+1,i4    'jmp     #scanret

                        'Init I/O registers and sync COGs' video circuits

                        mov     dira,reg_dira           'set pin directions
                        mov     dirb,reg_dirb
                        movi    frqa,#(pr / 5) << 2     'set pixel rate
                        mov     vcfg,reg_vcfg           'set video configuration
                        mov     vscl,#1                 'set video to reload on every pixel
                        waitcnt sync_cnt,colormask      'wait for start value in cnt, add ~1ms
                        movi    ctra,#%00001_110        'COGs in sync! enable PLLs now - NCOs locked!
                        waitcnt sync_cnt,#0             'wait ~1ms for PLLs to stabilize - PLLs locked!
                        mov     vscl,#100               'insure initial WAITVIDs lock cleanly

                        'Jump to main loop

                        jmp     #vsync                  'jump to vsync - WAITVIDs will now be locked!

                        'Data

d0s0                    long    1 << 9 + 1
d1                      long    1 << 10
main_ctr                long    main_size
scan_ctr                long    cols

i0                      waitvid x,scanbuff+0
i1                      shr     scanbuff+0,#8
i2                      mov     vscl,#hf
i3                      waitvid hvsync,#0
i4                      jmp     #scanret

reg_dira                long    0                       'set at runtime
reg_dirb                long    0                       'set at runtime
reg_vcfg                long    0                       'set at runtime
sync_cnt                long    0                       'set at runtime

                        'Directives

                        fit     scancode                'make sure initialization code and data fit
main_begin              org     maincode                'main code follows (gets moved into maincode)


' Main loop, display field - each COG alternately builds and displays four scan lines

vsync                   mov     x,#vs                   'do vertical sync lines
                        call    #blank_vsync

vb_lines                mov     x,#vb                   'do vertical back porch lines (# set at runtime)
                        call    #blank_vsync

                        mov     screen_ptr,screen_base  'reset screen pointer to upper-left character
                        mov     color_ptr,color_base    'reset color pointer to first row
                        mov     row,#0                  'reset row counter for cursor insertion
                        mov     fours,#rows * 4 / 2     'set number of 4-line builds for whole screen

                        'Build four scan lines into scanbuff

fourline                mov     font_ptr,font_quarter   'get address of appropriate font section
                        shl     font_ptr,#8+2           'shl 8 for 256 characters, 2 for indexing longs
                        add     font_ptr,font_base

                        movd    :pixa,#scanbuff-1       'reset scanbuff address (pre-decremented)

                        mov     y,#2                    'must build scanbuff in two sections because
                        mov     vscl,vscl_line2x        '..pixel counter is limited to twelve bits

:halfrow                waitvid underscore,#0           'output lows to let other COG drive VGA pins
                        mov     x,#cols/2               '..for 2 scan lines, ready for half a row

:column                 rdbyte  z,screen_ptr            'get character from screen memory
                        shl     z,#2                    'get chr into bits 9..2
                        add     z,font_ptr              'add font section address to point to 8*4 pixels
                        add     :pixa,d0                'increment scanbuff destination addresses
                        add     screen_ptr,#1           'increment screen memory address
:pixa                   rdlong  scanbuff,z              'read pixel long (8*4) into scanbuff
                        djnz    x,#:column              'another character in this half-row?

                        djnz    y,#:halfrow             'loop to do 2nd half-row, time for 2nd WAITVID

                        sub     screen_ptr,#cols        'back up to start of same row in screen memory

                        'Insert cursors into scanbuff

                        mov     z,#2                    'ready for two cursors

:cursor                 rdbyte  x,cursor_base           'x in range?
                        add     cursor_base,#1
                        cmp     x,#cols         wc

                        rdbyte  y,cursor_base           'y match?
                        add     cursor_base,#1
                        cmp     y,row           wz

                        rdbyte  y,cursor_base           'get cursor mode
                        add     cursor_base,#1

        if_nc_or_nz     jmp     #:nocursor              'if cursor not in scanbuff, no cursor

                        add     x,#scanbuff             'cursor in scanbuff, set scanbuff address
                        movd    :xor,x

                        test    y,#%010         wc      'get mode bits into flags
                        test    y,#%001         wz
        if_nc_and_z     jmp     #:nocursor              'if cursor disabled, no cursor

        if_c_and_z      test    slowbit,cnt     wc      'if blink mode, get blink state
        if_c_and_nz     test    fastbit,cnt     wc

                        test    y,#%100         wz      'get box or underscore cursor piece
        if_z            mov     x,longmask
        if_nz           mov     x,underscore
        if_nz           cmp     font_quarter,#3 wz      'if underscore, must be last font section

:xor    if_nc_and_z     xor     scanbuff,x              'conditionally xor cursor into scanbuff

:nocursor               djnz    z,#:cursor              'second cursor?

                        sub     cursor_base,#3*2        'restore cursor base

                        'Display four scan lines from scanbuff

                        rdword  x,color_ptr             'get color pattern for current row
                        and     x,colormask             'mask away hsync and vsync signal states
                        or      x,hv                    'insert inactive hsync and vsync states

                        mov     y,#4                    'ready for four scan lines

scanline                mov     vscl,vscl_chr           'set pixel rate for characters
                        jmp     #scancode               'jump to scanbuff display routine in scancode
scanret                 mov     vscl,#hs                'do horizontal sync pixels
                        waitvid hvsync,#1               '#1 makes hsync active
                        mov     vscl,#hb                'do horizontal back porch pixels
                        waitvid hvsync,#0               '#0 makes hsync inactive
                        shr     scanbuff+cols-1,#8      'shift last column's pixels right by 8
                        djnz    y,#scanline             'another scan line?

                        'Next group of four scan lines

                        add     font_quarter,#2         'if font_third + 2 => 4, subtract 4 (new row)
                        cmpsub  font_quarter,#4   wc    'c=0 for same row, c=1 for new row
        if_c            add     screen_ptr,#cols        'if new row, advance screen pointer
        if_c            add     color_ptr,#2            'if new row, advance color pointer
        if_c            add     row,#1                  'if new row, increment row counter
                        djnz    fours,#fourline         'another 4-line build/display?

                        'Visible section done, do vertical sync front porch lines

                        wrlong  longmask,par            'write -1 to refresh indicator

vf_lines                mov     x,#vf                   'do vertical front porch lines (# set at runtime)
                        call    #blank

                        jmp     #vsync                  'new field, loop to vsync

                        'Subroutine - do blank lines

blank_vsync             xor     hvsync,#$101            'flip vertical sync bits

blank                   mov     vscl,hx                 'do blank pixels
                        waitvid hvsync,#0
                        mov     vscl,#hf                'do horizontal front porch pixels
                        waitvid hvsync,#0
                        mov     vscl,#hs                'do horizontal sync pixels
                        waitvid hvsync,#1
                        mov     vscl,#hb                'do horizontal back porch pixels
                        waitvid hvsync,#0
                        djnz    x,#blank                'another line?
blank_ret
blank_vsync_ret         ret

                        'Data

screen_base             long    0                       'set at runtime (3 contiguous longs)
color_base              long    0                       'set at runtime
cursor_base             long    0                       'set at runtime

font_base               long    0                       'set at runtime
font_quarter            long    0                       'set at runtime

hx                      long    hp                      'visible pixels per scan line
vscl_line2x             long    (hp + hf + hs + hb) * 2 'total number of pixels per 2 scan lines
vscl_chr                long    2 << 12 + 16            '2 clocks per pixel and 16 pixels per set
colormask               long    $FCFC                   'mask to isolate R,G,B bits from H,V
longmask                long    $FFFFFFFF               'all bits set
slowbit                 long    1 << 25                 'cnt mask for slow cursor blink
fastbit                 long    1 << 24                 'cnt mask for fast cursor blink
underscore              long    $FFFF0000               'underscore cursor pattern
hv                      long    hv_inactive             '-H,-V states
hvsync                  long    hv_inactive ^ $200      '+/-H,-V states

                        'Uninitialized data

screen_ptr              res     1
color_ptr               res     1
font_ptr                res     1

x                       res     1
y                       res     1
z                       res     1

row                     res     1
fours                   res     1


' 8 x 16 font - characters 0..255
'
' Each long holds four scan lines of a single character. The longs are arranged into
' groups of 256 which represent all characters (0..255). There are four groups which
' each contain a vertical quarter of all characters. They are ordered top to bottom.

font  long
  long $828282fe,$020202fe,$10101010,$80808080,$20100804,$aac682fe,$20408000,$82824438 ' 1st quarter
  long $443e0408,$40201000,$000000fe,$10101000,$38541010,$04081000,$aac64438,$82824438
  long $828282fe,$92925438,$82824438,$82824438,$92925438,$20508800,$4444447c,$80808080
  long $284482fe,$38101010,$0242423c,$82824438,$929292fe,$828282fe,$828282fe,$929292fe
  long $00000000,$10101010,$00242424,$04044438,$1212fc10,$20448a04,$1422221c,$08103030
  long $08081020,$20201008,$54921000,$10101000,$00000000,$00000000,$00000000,$20408000
  long $a2c2827c,$10141810,$4080827c,$8080827c,$48506040,$3e0202fe,$02020478,$204082fe
  long $8282827c,$8282827c,$18000000,$18000000,$04081020,$7c000000,$40201008,$80848478
  long $aab28478,$82824438,$8484847e,$02028478,$8484443e,$020202fe,$020202fe,$02028478
  long $82828282,$1010107c,$202020f8,$12224282,$02020202,$92aac682,$928a8682,$82824438
  long $8282827e,$82824438,$8282827e,$0202827c,$101010fe,$82828282,$44828282,$82828282
  long $28448282,$28448282,$204080fe,$0404043c,$08040200,$2020203c,$54381000,$00000000
  long $20101818,$3c000000,$3a020202,$3c000000,$5c404040,$3c000000,$08084830,$5c000000
  long $3a020202,$18001000,$60004000,$22020202,$10101018,$6e000000,$3a000000,$3c000000
  long $3a000000,$5c000000,$3a000000,$3c000000,$3e080800,$42000000,$82000000,$82000000
  long $42000000,$42000000,$7e000000,$08080830,$00101010,$20202018,$000000fe,$24489224
  long $55aa55aa,$cc333300,$04020201,$20404080,$fcfefeff,$3f7f7fff,$07030301,$e0c0c080
  long $10204080,$08040201,$00000000,$00000000,$eeeeeeee,$ffffff00,$9e3e7e00,$f2f8fc00
  long $00000000,$00000000,$10101010,$10101010,$10101010,$10101010,$10101010,$10101010
  long $00000000,$10101010,$00000000,$101010f0,$1010101f,$000000ff,$fe282828,$36000000
  long $00000000,$04040604,$04040604,$100c100e,$00000000,$88700000,$00000000,$30080870
  long $00000000,$fc000000,$28100000,$100804fe,$40800000,$0c000000,$fe101010,$00001000
  long $48241208,$12244810,$202020e0,$82442810,$10284482,$3c20203c,$3c04043c,$3c24243c
  long $827c0000,$fe7c0000,$81c3ffff,$bdc3ffff,$7f3e1c08,$3e7f7722,$3e1c1c08,$7f7f1c1c
  long $00000000,$0f0f0f0f,$00000000,$0f0f0f0f,$00000000,$0f0f0f0f,$00000000,$0f0f0f0f
  long $f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff
  long $00000000,$0f0f0f0f,$00000000,$0f0f0f0f,$00000000,$0f0f0f0f,$00000000,$0f0f0f0f
  long $f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff
  long $00000000,$0f0f0f0f,$00000000,$0f0f0f0f,$00000000,$0f0f0f0f,$00000000,$0f0f0f0f
  long $f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff
  long $00000000,$0f0f0f0f,$00000000,$0f0f0f0f,$00000000,$0f0f0f0f,$00000000,$0f0f0f0f
  long $f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff,$f0f0f0f0,$ffffffff

  long $82828282,$02020202,$10101010,$80808080,$2010087c,$82c6aa92,$02060a12,$2828fe82 ' 2nd quarter
  long $80808088,$102040fe,$000000fe,$10385492,$38549210,$100804fe,$44c6aa92,$44828292
  long $828282fe,$448282f2,$549292f2,$5492929e,$4482829e,$02068a52,$44444444,$808080fe
  long $82442810,$10101038,$0010100c,$448282fe,$8282829e,$9292929e,$929292f2,$828282f2
  long $00000000,$10000010,$00000000,$8404040e,$7e90907c,$a2440810,$62629408,$00000000
  long $10080808,$10202020,$10925438,$101010fe,$18000000,$000000fe,$18000000,$02040810
  long $82868a92,$10101010,$02020438,$82808078,$40fe4244,$42808040,$8282827e,$08080810
  long $8282827c,$408080fc,$18000018,$18000018,$10080402,$00007c00,$10204080,$00101060
  long $04027aaa,$828282fe,$8484847c,$84020202,$44848484,$0202021e,$0202021e,$8482f202
  long $828282fe,$10101010,$22202020,$4222160a,$02020202,$82828292,$8282c2a2,$44828282
  long $0202027e,$44a29282,$4222127e,$8280807c,$10101010,$82828282,$10282844,$c6aa9292
  long $82442810,$10101010,$02040810,$04040404,$80402010,$20202020,$10101092,$00000000
  long $00000000,$42427c40,$46424246,$42020242,$62424262,$02027e42,$0808083e,$62424262
  long $42424246,$10101010,$40404040,$22160a12,$10101010,$92929292,$42424246,$42424242
  long $46424246,$62424262,$02020246,$42300c42,$48080808,$62424242,$28448282,$92929292
  long $24181824,$62424242,$04081020,$08080804,$10101000,$20202040,$00000000,$92244892
  long $55aa55aa,$cc3333cc,$10080804,$08101020,$f0f8f8fc,$0f1f1f3f,$1f0f0f07,$f8f0f0e0
  long $01020408,$80402010,$03000000,$c0000000,$eeeeeeee,$ffff0000,$f8f2e6ce,$3e9ecee6
  long $f0000000,$1f000000,$f0101010,$1f101010,$10101010,$f0101010,$ff101010,$1f101010
  long $ff000000,$ff101010,$ff000000,$1f101010,$f0101010,$00000000,$28fe2828,$00364949
  long $00000000,$5060400e,$4090600e,$5060400e,$5222528c,$88887888,$28448200,$48487040
  long $44444444,$2828282a,$82444428,$04081020,$fe1010fe,$0c006092,$00101010,$100000fe
  long $12244890,$48241209,$302a2420,$00000000,$107c10fe,$bd7e3c18,$bd7e3c18,$b5adff18
  long $82828282,$fefefefe,$c3818181,$c3bdbdbd,$08082a7f,$081c1c3e,$1c1c3e7f,$3e080808
  long $00000000,$0000000f,$0f0f0f00,$0f0f0f0f,$00000000,$0000000f,$0f0f0f00,$0f0f0f0f
  long $000000f0,$000000ff,$0f0f0ff0,$0f0f0fff,$000000f0,$000000ff,$0f0f0ff0,$0f0f0fff
  long $f0f0f000,$f0f0f00f,$ffffff00,$ffffff0f,$f0f0f000,$f0f0f00f,$ffffff00,$ffffff0f
  long $f0f0f0f0,$f0f0f0ff,$fffffff0,$ffffffff,$f0f0f0f0,$f0f0f0ff,$fffffff0,$ffffffff
  long $00000000,$0000000f,$0f0f0f00,$0f0f0f0f,$00000000,$0000000f,$0f0f0f00,$0f0f0f0f
  long $000000f0,$000000ff,$0f0f0ff0,$0f0f0fff,$000000f0,$000000ff,$0f0f0ff0,$0f0f0fff
  long $f0f0f000,$f0f0f00f,$ffffff00,$ffffff0f,$f0f0f000,$f0f0f00f,$ffffff00,$ffffff0f
  long $f0f0f0f0,$f0f0f0ff,$fffffff0,$ffffffff,$f0f0f0f0,$f0f0f0ff,$fffffff0,$ffffffff

  long $000000fe,$00000002,$000000fe,$000000fe,$00000040,$000000fe,$00000000,$000000ee ' 3rd quarter
  long $00000080,$00000000,$000000fe,$00000000,$00000010,$00000000,$00000038,$00000038
  long $000000fe,$00000038,$00000038,$00000038,$00000038,$00000000,$000000c6,$00000080
  long $000000fe,$00000010,$00000010,$00000038,$000000fe,$000000fe,$000000fe,$000000fe
  long $00000000,$00000010,$00000000,$0000007e,$00000010,$00000040,$0000009c,$00000000
  long $00000020,$00000008,$00000000,$00000000,$00040818,$00000000,$00000018,$00000000
  long $0000007c,$0000007c,$000000fe,$0000007c,$00000040,$0000003c,$0000007c,$00000008
  long $0000007c,$0000003c,$00000018,$00040818,$00000020,$00000000,$00000008,$00000010
  long $00000078,$00000082,$0000007e,$00000078,$0000003e,$000000fe,$00000002,$00000078
  long $00000082,$0000007c,$0000001c,$00000082,$000000fe,$00000082,$00000082,$00000038
  long $00000002,$000000b8,$00000082,$0000007c,$00000010,$0000007c,$00000010,$00000082
  long $00000082,$00000010,$000000fe,$0000003c,$00000000,$0000003c,$00000000,$000000fe
  long $00000000,$000000bc,$0000003a,$0000003c,$0000005c,$0000003c,$00000008,$3c42405c
  long $00000042,$00000038,$38444040,$00000042,$00000038,$00000092,$00000042,$0000003c
  long $0202023a,$4040405c,$00000002,$0000003c,$00000030,$0000005c,$00000010,$0000006c
  long $00000042,$3c42405c,$0000007e,$00000030,$00000000,$00000018,$00000000,$00000048
  long $55aa55aa,$000000cc,$40202010,$02040408,$c0e0e0f0,$0307070f,$7f3f3f1f,$fefcfcf8
  long $00000000,$00000000,$20100804,$04081020,$000000ee,$000000ff,$000000fc,$0000007e
  long $10101010,$10101010,$00000000,$00000000,$10101010,$10101010,$10101010,$10101010
  long $00000000,$00000000,$10101010,$00000000,$00000000,$00000000,$00002828,$00000000
  long $00000000,$000040f8,$0000f020,$000040f8,$0000008c,$08080878,$10282810,$00000030
  long $040404bc,$00000028,$000000fe,$000000fe,$00000204,$00006092,$0000fe00,$00000000
  long $00000008,$00000010,$00000020,$00000000,$00000010,$143cbdbd,$283cbdbd,$243cb5ad
  long $0000007c,$0000007c,$0000ffff,$0000ffff,$0000003e,$00000000,$00000008,$00000000
  long $00000000,$00000000,$00000f0f,$00000f0f,$0f0f0000,$0f0f0000,$0f0f0f0f,$0f0f0f0f
  long $00000000,$00000000,$00000f0f,$00000f0f,$0f0f0000,$0f0f0000,$0f0f0f0f,$0f0f0f0f
  long $0000f0f0,$0000f0f0,$0000ffff,$0000ffff,$0f0ff0f0,$0f0ff0f0,$0f0fffff,$0f0fffff
  long $0000f0f0,$0000f0f0,$0000ffff,$0000ffff,$0f0ff0f0,$0f0ff0f0,$0f0fffff,$0f0fffff
  long $f0f00000,$f0f00000,$f0f00f0f,$f0f00f0f,$ffff0000,$ffff0000,$ffff0f0f,$ffff0f0f
  long $f0f00000,$f0f00000,$f0f00f0f,$f0f00f0f,$ffff0000,$ffff0000,$ffff0f0f,$ffff0f0f
  long $f0f0f0f0,$f0f0f0f0,$f0f0ffff,$f0f0ffff,$fffff0f0,$fffff0f0,$ffffffff,$ffffffff
  long $f0f0f0f0,$f0f0f0f0,$f0f0ffff,$f0f0ffff,$fffff0f0,$fffff0f0,$ffffffff,$ffffffff

  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000 ' 4th quarter
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $55aa55aa,$00000000,$80808040,$01010102,$808080c0,$01010103,$ffffff7f,$fffffffe
  long $00000000,$00000000,$80808040,$01010102,$00000000,$00000000,$00000000,$00000000
  long $10101010,$10101010,$00000000,$00000000,$10101010,$10101010,$10101010,$10101010
  long $00000000,$00000000,$10101010,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000004,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$00000000,$63424222,$c6424244,$66242424
  long $00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000,$00000000
  long $00000000,$00000000,$00000000,$00000000,$0f0f0f0f,$0f0f0f0f,$0f0f0f0f,$0f0f0f0f
  long $00000000,$00000000,$00000000,$00000000,$0f0f0f0f,$0f0f0f0f,$0f0f0f0f,$0f0f0f0f
  long $00000000,$00000000,$00000000,$00000000,$0f0f0f0f,$0f0f0f0f,$0f0f0f0f,$0f0f0f0f
  long $00000000,$00000000,$00000000,$00000000,$0f0f0f0f,$0f0f0f0f,$0f0f0f0f,$0f0f0f0f
  long $f0f0f0f0,$f0f0f0f0,$f0f0f0f0,$f0f0f0f0,$ffffffff,$ffffffff,$ffffffff,$ffffffff
  long $f0f0f0f0,$f0f0f0f0,$f0f0f0f0,$f0f0f0f0,$ffffffff,$ffffffff,$ffffffff,$ffffffff
  long $f0f0f0f0,$f0f0f0f0,$f0f0f0f0,$f0f0f0f0,$ffffffff,$ffffffff,$ffffffff,$ffffffff
  long $f0f0f0f0,$f0f0f0f0,$f0f0f0f0,$f0f0f0f0,$ffffffff,$ffffffff,$ffffffff,$ffffffff

''      MIT LICENSE
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
